/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>

#include "memory.h"
#include "cpu.h"
#include "hardware.h"
#include "log.h"
#include "lcd.h"
#include "usb.h"
#include "flash.h"

unsigned int tick;

int tick_before(unsigned int a, unsigned int b) {
  return (unsigned int)(b - a) < (1 << 30);
}

static uint32 gpio_read(struct hardware *hardware, uint32 adr) {
  int n = (adr >> 6) & 7;
  int r = adr & 0x3f;
  int i;
  uint32 res;

  switch (r) {
  case 0:
    logmsg(LOG_GPIO, 20, "gpio(%d): read from register %x\n", n, r);
    res = 0;
    for (i = 31; i >= 0; i--) {
      struct gpio_handler *h = &(hardware->gpio_handler[n][i]);
      if (h->get) {
	res = (res << 1) | (*(h->get))(h->data, n, i);
      } else
	res = (res << 1) | 1;
    }
    return res;
  default:
    logmsg(LOG_GPIO, 5, "gpio(%d): read from register %x\n", n, r);
    return ~0;
  }
}

static void gpio_notify_handlers(struct hardware *hardware,
				 int port, uint32 old, uint32 new) {
  int i;
  for (i = 0; i < 32; i++)
    if ((old ^ new) & (1 << i)) {
      struct gpio_handler *h = &(hardware->gpio_handler[port][i]);
      if (h->set)
	(*h->set)(h->data, port, i, (new >> i) & 1);
    }
      
}

int gpio_get_bit(struct hardware *hardware, int port, int bit) {
  return (hardware->gpio_out[port] >> bit) & 1;
}

void gpio_install_handler(struct hardware *hardware, int port, int bit,
			  void *data,
			  int (*get)(void *, int, int),
			  void (*set)(void *, int, int, int)) {
  hardware->gpio_handler[port][bit].data = data;
  hardware->gpio_handler[port][bit].get = get;
  hardware->gpio_handler[port][bit].set = set;
}

static void gpio_direction_changed(struct hardware *hardware, int port,
				   uint32 mask, uint32 new) {
  int i;
  for (i = 0; i < 32; i++)
    if (mask & (1 << i))
      logmsg(LOG_GPIO, 10, "gpio: bit %d/%d set to %s\n",
	     port, i, (new & (1 << i)) ? "output" : "input");
}

static void gpio_write(struct hardware *hardware, uint32 adr, uint32 val) {
  int n = (adr >> 6) & 7;
  int r = adr & 0x3c;
  uint32 old;

  switch (r) {
  case 0x14:
    old = hardware->gpio_out[n];
    hardware->gpio_out[n] |= val;
    gpio_notify_handlers(hardware,
			 n, old, hardware->gpio_out[n]);
    break;
  case 0x18:
    old = hardware->gpio_out[n];
    hardware->gpio_out[n] &= ~val;
    gpio_notify_handlers(hardware,
			 n, old, hardware->gpio_out[n]);
    break;
  case 0x24:
    hardware->gpio_dir[n] |= val;
    gpio_direction_changed(hardware, n,
			   val, hardware->gpio_dir[n]);
    break;
  case 0x28:
    hardware->gpio_dir[n] &= ~val;
    gpio_direction_changed(hardware, n,
			   val, hardware->gpio_dir[n]);
    break;
  default:
    logmsg(LOG_GPIO, 5, "gpio(%d): write %x to register %x\n", n, val, r);
  }
}

static int flash_ce[4] = {0x20, 0x02, 0x10, 0x08};

static int selected_flash(struct hardware *hardware) {
  int i, r;
  uint32 p5 = hardware->gpio_out[5];
  r = -1;
  for (i = 0; i < 4; i++)
    if (!(p5 & flash_ce[i])) {
      if (r > 0)
	logmsg(LOG_FLASH, 5, "flash: more than one chip selected\n");
      else
	r = i;
    }
  return r;
}

static int flash_get_busy(void *data, int port, int bit) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1)
    return 1;

  flash = hardware->flash[chip];
  if (!flash)
    return 1;

  if (!flash->busy)
    return 1;

  if (tick_before(flash->ready_tick, tick)) {
    flash->busy = 0;
    return 1;
  }

  return 0;
}

static uint32 prog_flash_read_half_word(void *data, uint32 adr,
					struct profile *profile,
					uint32 req_adr) {
  struct prog_flash *flash = data;
  logmsg(LOG_FW_FLASH, 10, "prog_flash: read from %x\n", adr);
  adr = adr & 0xffffe;

  switch (flash->state) {
  case 0:
    return flash->content[adr] | (flash->content[adr + 1] << 8);

  case 0x90:
    switch (adr) {
    case 0:
      return 0xbf;
    case 2:
      return 0x2781;
    default:
      return 0xffff;
    }
  default:
    return 0xffff;
  }
}

static uint8 prog_flash_read_byte(void *data, uint32 adr,
				  struct profile *profile, uint32 req_adr) {
  uint8 res;
  if (adr & 1)
    res = prog_flash_read_half_word(data, adr & ~1, profile, req_adr) >> 8;
  else
    res = prog_flash_read_half_word(data, adr, profile, req_adr) & 0xff;
  logmsg(LOG_FW_FLASH, 10, "prog_flash: read byte %x\n", res);
  return res;
}

static uint32 prog_flash_read_word(void *data, uint32 adr,
				   struct profile *profile, uint32 req_adr) {
  return prog_flash_read_half_word(data, adr & ~3, profile, req_adr)
    | (prog_flash_read_half_word(data, (adr & ~3) + 2,
				 profile, req_adr) << 16);
}

static int prog_flash_update(struct prog_flash *flash, int start, int size) {
  if (fseek(flash->fp, start, SEEK_SET))
    return -1;
  fwrite(flash->content + start, 1, size, flash->fp);
  return 0;
}

static void prog_flash_sector_erase(struct prog_flash *flash, int sector) {
  sector = sector & 0xff;
  logmsg(LOG_FW_FLASH, 5, "prog_flash: sector %x erase\n", sector);
  memset(flash->content + sector * 4096, 0xff, 4096);
  prog_flash_update(flash, sector * 4096, 4096);
}

static void prog_flash_block_erase(struct prog_flash *flash, int block) {
  block = block & 0xf;
  logmsg(LOG_FW_FLASH, 5, "prog_flash: block %x erase\n", block);
  memset(flash->content + block * 64 * 1024, 0xff, 64 * 1024);
  prog_flash_update(flash, block * 64 * 1024, 64 * 1024);
}

static void prog_flash_chip_erase(struct prog_flash *flash) {
  logmsg(LOG_FW_FLASH, 5, "prog_flash: chip erase\n");
  memset(flash->content, 0xff, 1024 * 1024);
  prog_flash_update(flash, 0, 1024 * 1024);
}

static void prog_flash_write_half_word(void *data, uint32 adr, uint32 val,
				       struct profile *profile,
				       uint32 req_adr) {
  struct prog_flash *flash = data;
  int adr15 = (adr & 0xffff) >> 1;
  int val8 = val & 0xff;

  logmsg(LOG_FW_FLASH, 10, "prog_flash: write %x to %x\n", val, adr);

  switch (flash->state) {
  case 0:
    if (adr15 == 0x5555 && val8 == 0xaa)
      flash->state = 1;
    break;

  case 1:
    if (adr15 == 0x2aaa && val8 == 0x55)
      flash->state = 2;
    else
      flash->state = 0;
    break;

  case 2:
    if (adr15 == 0x5555) {
      switch (val8) {
      case 0xa0:
      case 0x80:
      case 0x90:
      case 0x98:
	flash->state = val & 0xff;
	break;
      default:
	flash->state = 0;
	break;
      }
    } else
      flash->state = 0;
    break;

  case 0x80:
    /* erase mode */
    if (adr15 == 0x5555 && val8 == 0xaa)
      flash->state = 0x81;
    else
      flash->state = 0;
    break;

  case 0x81:
    if (adr15 == 0x2aaa && val8 == 0x55)
      flash->state = 0x82;
    else
      flash->state = 0;
    break;

  case 0x82:
    switch (val8) {
    case 0x30:
      prog_flash_sector_erase(flash, adr >> 12);
      flash->state = 0;
      break;
    case 0x50:
      prog_flash_block_erase(flash, adr >> 16);
      flash->state = 0;
      break;
    case 0x10:
      if (adr15 == 0x5555)
	prog_flash_chip_erase(flash);
      flash->state = 0;
      break;
    default:
      flash->state = 0;
      break;
    }
    break;
  case 0x90:
    /* software id mode */
    if (adr15 == 0x5555 && val8 == 0xf0)
      flash->state = 0;
    break;

  case 0x98:
    /* TODO: CFI query */
    flash->state = 0;
    break;

  case 0xa0:
    adr = adr & 0xffffe;
    flash->content[adr] = val & 0xff;
    flash->content[adr + 1] = val >> 8;
    prog_flash_update(flash, adr, 2);
    flash->state = 0;
    break;

  default:
    flash->state = 0;
    break;
  }
}

static uint32 read_803(struct hardware *hardware, uint32 adr) {
  logmsg(LOG_MEMORY, 5, "irq: read from %x\n", adr);

  if ((adr & 0xfff) < 0x500)
    return hardware->irq_reg[(adr & 0xfff) >> 2];
  else
    return ~0;
}

static void write_803(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "irq: write %x to %x\n", val, adr);

  if ((adr & 0xfff) < 0x500)
    hardware->irq_reg[(adr & 0xfff) >> 2] = val;
}

int check_irq(void *data) {
  struct hardware *hardware = data;
  int i;
  
  if (hardware->irq_request) {
    for (i = 0; i < 32; i++)
      if (hardware->irq_request & (1 << i)) {
	hardware->irq_request &= ~(1 << i);
	hardware->irq_reg[0x40] = i * 8;
	return 1;
      }
  }
  return 0;
}

void set_irq_request(struct hardware *hardware, int i) {
  if (hardware->irq_reg[0x100 + i] & 0x10000)
    hardware->irq_request |= (1 << i);
}

/* Timer implementation is quite lame, but that is enough.  More
   advanced event handling (removing, changing time) would be
   required. */

static void timer_handle(struct pnx0101_timer *timer) {
  if (timer->ctrl & 0x80) {
    unsigned int delta = tick - timer->last_tick;
    if (timer->counter < delta) {
      timer->counter = timer->reload;
      set_irq_request(timer->hardware, 5 + timer->index);
    } else
      timer->counter -= delta;
  }
  timer->last_tick = tick;
}

static void timer_add_event(struct pnx0101_timer *timer) {
  if (!timer->active && (timer->ctrl & 0x80)) {
    timer->active = 1;
    timer->event.tick = tick + timer->counter + 1;
    add_event(&(timer->event));
  }
}

static void timer_event(struct event *e) {
  struct pnx0101_timer *timer = e->data;
  timer_handle(timer);
  if (timer->ctrl & 0x80) {
    e->tick = tick + timer->counter + 1;
    add_event(e);
  } else
    timer->active = 0;
}

static uint32 timer_read(struct hardware *hardware, uint32 addr) {
  struct pnx0101_timer *timer = &(hardware->timer[(addr & 0x400) ? 1 : 0]);
  timer_handle(timer);
  switch (addr & 0xf) {
  case 8:
    logmsg(LOG_MEMORY, 10, "timer(%d): read from ctrl\n", timer->index);
    return timer->ctrl;
  default:
    logmsg(LOG_MEMORY, 5, "timer(%d): read from %x\n", timer->index, addr);
  }
  return ~0;
}

static void timer_write(struct hardware *hardware,
			uint32 addr, uint32 val) {
  struct pnx0101_timer *timer = &(hardware->timer[(addr & 0x400) ? 1 : 0]);
  timer_handle(timer);
  switch (addr & 0xf) {
  case 0:
    logmsg(LOG_MEMORY, 10, "timer(%d): write %x to reload\n",
	   timer->index, val);
    timer->reload = timer->counter = val * 2;
    timer_add_event(timer);
    break;
  case 8:
    logmsg(LOG_MEMORY, 10, "timer(%d): write %x to ctrl\n",
	   timer->index, val);
    timer->ctrl = val;
    timer_add_event(timer);
    break;
  default:
    logmsg(LOG_MEMORY, 5, "timer(%d): write %x to %x (pc=%8x)\n",
	   timer->index, val, addr, cpu->reg[15] - 4);
  }
}

static uint32 adc_read(struct hardware *hardware, uint32 adr) {
  switch (adr) {
  case 0: /* most of buttons */
    if (ui_get_key(2)) return 0x280;
    if (ui_get_key(3)) return 0x300;
    if (ui_get_key(4)) return 0x100;
    if (ui_get_key(5)) return 0x20;
    if (ui_get_key(6)) return 0x60;
    if (ui_get_key(7)) return 0x1a0;
    if (ui_get_key(8)) return 0x200;
    return 0x3ff;
  case 4: /* battery */
    return 450;
  case 8:
    if (ui_get_key(1)) return 0;
    return 0x3ff;
  default:
    logmsg(LOG_ADC, 5, "adc: read from %x\n", adr);
  }
  return ~0;
}

static void adc_write(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_ADC, 5, "adc: write %x to %x\n", val, adr);
}

static uint32 audio_read(struct audio *audio, uint32 adr) {
  switch (adr) {
  case 0x384:
    logmsg(LOG_MEMORY, 5, "audio: siocr read\n");
    return audio->siocr;
  case 0x398:
    logmsg(LOG_MEMORY, 5, "audio: dacctrl read\n");
    return audio->dacctrl;
  case 0x3a0:
    logmsg(LOG_MEMORY, 5, "audio: dacset read\n");
    return audio->dacset;
  default:
    logmsg(LOG_MEMORY, 5, "audio: read from %x\n", adr);
  }
  return ~0;
}

static void audio_output(struct audio *audio, int channel, uint32 val) {
  if (audio->out_file[channel]) {
    fputc((val >> 8) & 0xff, audio->out_file[channel]);
    fputc((val >> 16) & 0xff, audio->out_file[channel]);
  }
}

static void audio_write(struct audio *audio,
			uint32 addr, uint32 val) {
  switch (addr) {
  case 0x280:
    audio_output(audio, AUDIO_LEFT, val << 8);
    break;
  case 0x284:
    audio_output(audio, AUDIO_RIGHT, val << 8);
    break;
  case 0x288:
    audio_output(audio, AUDIO_LEFT, val);
    break;
  case 0x28c:
    audio_output(audio, AUDIO_RIGHT, val);
    break;
  case 0x2a0:
    audio_output(audio, AUDIO_LEFT, (val & 0xffff) << 8);
    audio_output(audio, AUDIO_LEFT, (val >> 8) & 0xffff00);
    break;
  case 0x2c0:
    audio_output(audio, AUDIO_RIGHT, (val & 0xffff) << 8);
    audio_output(audio, AUDIO_RIGHT, (val >> 8) & 0xffff00);
    break;
  case 0x2e0:
    audio_output(audio, AUDIO_LEFT, (val & 0xffff) << 8);
    audio_output(audio, AUDIO_RIGHT, (val >> 8) & 0xffff00);
    break;
  case 0x384:
    audio->siocr = val;
    logmsg(LOG_MEMORY, 5, "audio: siocr set to %x (dai_oe=%d)\n",
	   val, bits(val, 7, 1));
    break;
  case 0x398:
    audio->dacctrl = val;
    logmsg(LOG_MEMORY, 5, "audio: dacctrl set to %x "
	   "(rgain=%x, lgain=%x, deemph=%x, smute=%x, mode2fs=%x, "
	   "rolloff=%x, pslow=%x, ddac_pd=%x, ddac_inv=%x, sildet_t=%x, "
	   "ensildet=%x)\n", val, bits(val, 0, 8), bits(val, 8, 8),
	   bits(val, 16, 3), bits(val, 19, 1), bits(val, 20, 2),
	   bits(val, 22, 2), bits(val, 24, 1), bits(val, 25, 1),
	   bits(val, 26, 1), bits(val, 27, 2), bits(val, 29, 1));
    break;
  case 0x3a0:
    audio->dacset = val;
    logmsg(LOG_MEMORY, 5, "audio: dacset set to %x "
	   "(rdynpon=%d, ldynpon=%d, lbi_dwa=%d, rbi_dwa=%d)\n", val,
	   bits(val, 8, 1), bits(val, 9, 1), bits(val, 10, 1),
	   bits(val, 11, 1));
    break;
  default:
    logmsg(LOG_MEMORY, 5, "audio: write %x to %x (pc=%8x)\n",
	   val, addr, cpu->reg[15] - 4);
  }
}

static uint32 clocks_read(struct clocks *clocks, uint32 adr) {
  if (adr >= 0 && adr < 0x30) {
    logmsg(LOG_CGU, 5, "cgu: read from base_scr %x\n", adr >> 2);
    return clocks->base_scr[adr >> 2];
  } else if (adr >= 0x30 && adr < 0x60) {
    logmsg(LOG_CGU, 5, "cgu: read from base_fs1 %x\n", (adr - 0x30) >> 2);
    return clocks->base_fs1[(adr - 0x30) >> 2];
  } else if (adr >= 0x60 && adr < 0x90) {
    logmsg(LOG_CGU, 5, "cgu: read from base_fs2 %x\n", (adr - 0x60) >> 2);
    return clocks->base_fs2[(adr - 0x60) >> 2];
  } else if (adr >= 0x90 && adr < 0xc0) {
    logmsg(LOG_CGU, 5, "cgu: read from base_ssr %x\n", (adr - 0x90) >> 2);
    return clocks->base_ssr[(adr - 0x90) >> 2];
  } else if (adr >= 0xc0 && adr <= 0xc0 + 72 * 4) {
    logmsg(LOG_CGU, 5, "cgu: read from clk_pcr %x\n", (adr - 0xc0) >> 2);
    return clocks->clk_pcr[(adr - 0xc0) >> 2];
  } else if (adr >= 0x308 && adr <= 0x308 + 66 * 4) {
    logmsg(LOG_CGU, 5, "cgu: read from clk_esr %x\n", (adr - 0x308) >> 2);
    return clocks->clk_esr[(adr - 0x308) >> 2];
  } else if (adr >= 0x414 && adr <= 0x414 + 2 * 4) {
    logmsg(LOG_CGU, 5, "cgu: read from base_bcr %x\n", (adr - 0x414) >> 2);
    return clocks->base_bcr[(adr - 0x414) >> 2];
  } else if (adr >= 0x420 && adr <= 0x420 + 17 * 4) {
    logmsg(LOG_CGU, 5, "cgu: read from base_fdc %x\n", (adr - 0x420) >> 2);
    return clocks->base_fdc[(adr - 0x420) >> 2];
  } else if (adr >= 0xcac && adr <= 0xce0) {
    logmsg(LOG_CGU, 5, "cgu: read from regcac %x\n", (adr - 0xcac));
    return ~0;
  } else if (adr >= 0xce4 && adr <= 0xcfc) {
    logmsg(LOG_CGU, 5, "cgu: read from regce4 %x\n", (adr - 0xce4));
    return ~0;
  } else {
    logmsg(LOG_CGU, 5, "cgu: read from %x\n", adr);
    return -1;
  }
}

static void clocks_write(struct clocks *clocks,
			uint32 adr, uint32 val) {
  if (adr >= 0 && adr < 0x30) {
    logmsg(LOG_CGU, 5, "cgu: write %x to base_scr %x\n", val, adr >> 2);
    clocks->base_scr[adr >> 2] = val;
  } else if (adr >= 0x30 && adr < 0x60) {
    logmsg(LOG_CGU, 5, "cgu: write %x to base_fs1 %x\n",
	   val, (adr - 0x30) >> 2);
    clocks->base_fs1[(adr - 0x30) >> 2] = val;
  } else if (adr >= 0x60 && adr < 0x90) {
    logmsg(LOG_CGU, 5, "cgu: write %x to base_fs2 %x\n",
	   val, (adr - 0x60) >> 2);
    clocks->base_fs2[(adr - 0x60) >> 2] = val;
  } else if (adr >= 0x90 && adr < 0xc0) {
    logmsg(LOG_CGU, 5, "cgu: write %x to base_ssr %x\n",
	   val, (adr - 0x90) >> 2);
    clocks->base_ssr[(adr - 0x90) >> 2] = val;
  } else if (adr >= 0xc0 && adr <= 0xc0 + 72 * 4) {
    logmsg(LOG_CGU, 5, "cgu: write %x to clk_pcr %x\n",
	   val, (adr - 0xc0) >> 2);
    clocks->clk_pcr[(adr - 0xc0) >> 2] = val;
  } else if (adr >= 0x308 && adr <= 0x308 + 66 * 4) {
    logmsg(LOG_CGU, 5, "cgu: write %x to clk_esr %x\n",
	   val, (adr - 0x308) >> 2);
    clocks->clk_esr[(adr - 0x308) >> 2] = val;
  } else if (adr >= 0x414 && adr <= 0x414 + 2 * 4) {
    logmsg(LOG_CGU, 5, "cgu: write %x to base_bcr %x\n",
	   val, (adr - 0x414) >> 2);
    clocks->base_bcr[(adr - 0x414) >> 2] = val;
  } else if (adr >= 0x420 && adr <= 0x420 + 17 * 4) {
    int n = (adr - 0x420) >> 2;
    logmsg(LOG_CGU, 5, "cgu: write %x (%x %x %x %x) to base_fdc %x\n",
	   val,
	   val >> ((n == 0x10) ? 13 : 11),
	   (val >> 3) & ((n == 0x10) ? 0x3ff : 0xff),
	   (val >> 2) & 1,
	   val & 1,
	   n);
    clocks->base_fdc[(adr - 0x420) >> 2] = val;
  } else if (adr >= 0xcac && adr <= 0xce0) {
    logmsg(LOG_CGU, 5, "clocks: write %x to regcac %x\n", val, (adr - 0xcac));
  } else if (adr >= 0xce4 && adr <= 0xcfc) {
    logmsg(LOG_CGU, 5, "clocks: write %x to regce4 %x\n", val, (adr - 0xce4));
  } else {
    logmsg(LOG_CGU, 5, "clocks: write %x to %x\n", val, adr);
  }
}

static struct base_clock {
  int first_div;
  int n_div;
  int first_clock;
  int last_clock;
} base_clock[12] =
  {
    {0, 7, 0, 27},
    {7, 2, 28, 36},
    {9, 1, 37, 40},
    {10, 1, 41, 56},
    {0, 0, 57, 57},
    {0, 0, 58, 58},
    {0, 0, 59, 59},
    {0, 0, 60, 60},
    {11, 1, 61, 61},
    {12, 6, 62, 70},
    {0, 0, 71, 71},
    {0, 0, 72, 72}
  };

static int clocks_clock_base_clock(int clock) {
  int i;
  for (i = 0; i < 12; i++)
    if (base_clock[i].last_clock >= clock)
      return i;
  return -1;
}

static void clocks_log_state(struct clocks *clocks) {
  int i, esr;
  esr = 0;
  for (i = 0; i < 73; i++) {
    int base_id;
    base_id = clocks_clock_base_clock(i);
    if (base_clock[base_id].n_div > 0) {
      int div = base_clock[base_id].first_div + (clocks->clk_esr[esr] >> 1);
      int divcfg = clocks->base_fdc[div];
      int madd, msub;
      if (div != 16) {
	madd = (divcfg >> 3) & 0xff;
	msub = 256 - ((divcfg >> 11) & 0xff);
      } else {
	madd = (divcfg >> 3) & 0x3ff;
	msub = 1024 - ((divcfg >> 13) & 0x3ff);
      }	
      if (clocks->clk_esr[esr] & 1)
	logmsg(LOG_CGU, 5, "cgu: clock %2d: divider %2d "
	       "(run:%d strtch:%d n:%d m:%d), input %d\n",
	       i, div, divcfg & 1, (divcfg >> 2) & 1, msub, madd + msub,
	       (clocks->base_scr[base_id] & 1) ?
	       clocks->base_fs1[base_id] : clocks->base_fs2[base_id]);
      else
	logmsg(LOG_CGU, 5, "cgu: clock %2d: no divider, input %d\n",
	       i, (clocks->base_scr[base_id] & 1) ?
	       clocks->base_fs1[base_id] : clocks->base_fs2[base_id]);
      esr++;
    } else 
      logmsg(LOG_CGU, 5, "cgu: clock %2d: input %d\n",
	  i, (clocks->base_scr[base_id] & 1) ?
	  clocks->base_fs1[base_id] : clocks->base_fs2[base_id]);
  }
}

static uint32 i2c_read(struct hardware *hardware, uint32 adr) {
  switch (adr & 0xff) {
  case 4:
    logmsg(LOG_I2C, 5, "i2c: status read\n", adr);
    return 0;
  default:
    logmsg(LOG_I2C, 5, "i2c: read from %x\n", adr);
  }
  return ~0;
}

static void i2c_write(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_I2C, 5, "i2c: write %x to %x\n", val, adr);
}

static uint8 *cache_read(struct mmu_block *b, uint32 adr,
			 struct profile *profile, uint32 req_adr) {
  struct mmu_cache_line *l = &(b->cache->line[(adr >> 5) & 0x7f]);
  uint32 tag_addr = adr & ~0x1f;
  uint32 phys_adr;
  uint8* data;
  int i;
  for (i = 0; i < 2; i++)
    if (l->addr[i] == tag_addr) {
      profile_count(profile, req_adr, PROFILE_CTR_CACHE_HITS, 1);
      l->lru = i ^ 1;
      return l->data[i] + (adr & 0x1f);
    }
  profile_count(profile, req_adr, PROFILE_CTR_CACHE_MISSES, 1);
  data = l->data[l->lru];
  if (l->dirty[l->lru]) {
    logmsg(LOG_MEMORY, 15, "cache flush %x\n", l->addr[l->lru]);
    profile_count(profile, req_adr, PROFILE_CTR_CACHE_FLUSHES, 1);
    phys_adr = b->cache->block[l->addr[l->lru] >> 21].phys_adr
      + (l->addr[l->lru] & ((1 << 21) - 1) & ~0x1f);
    for (i = 0; i < 32; i += 4)
      memory_write_word(b->memory, phys_adr + i, get_word(data + i),
			profile, req_adr);
    l->dirty[l->lru] = 0;
  }

  phys_adr = b->phys_adr + (adr & ((1 << 21) - 1) & ~0x1f);
  l->addr[l->lru] = tag_addr;
  logmsg(LOG_MEMORY, 15, "cache fetch %x\n", l->addr[l->lru]);
  for (i = 0; i < 32; i += 4)
    put_word(data + i,
	     memory_read_word(b->memory, phys_adr + i, profile, req_adr));
  l->lru ^= 1;
  return data + (adr & 0x1f);
}

static uint8 *cache_write(struct mmu_block *b, uint32 adr,
			  struct profile *profile, uint32 req_adr) {
  struct mmu_cache_line *l = &(b->cache->line[(adr >> 5) & 0x7f]);
  uint32 tag_addr = adr & ~0x1f;
  int i;

  if (!(b->cache->settings & 6) || !b->cache_enabled)
    return NULL;

  for (i = 0; i < 2; i++)
    if (l->addr[i] == tag_addr) {
      profile_count(profile, req_adr, PROFILE_CTR_CACHE_HITS, 1);
      l->dirty[i] = 1;
      l->lru = i ^ 1;
      return l->data[i] + (adr & 0x1f);
    }
  profile_count(profile, req_adr, PROFILE_CTR_CACHE_MISSES, 1);
  return NULL;
}

static uint8 mmu_block_read(void *data, uint32 adr,
			    struct profile *profile, uint32 req_adr) {
  struct mmu_block *b = data;
  if ((b->cache->settings & 6) && b->cache_enabled) {
    uint8 *p = cache_read(b, adr, profile, req_adr);
    return *p;
  } else
    return memory_read(b->memory, b->phys_adr + (adr & ((1 << 21) - 1)),
		       profile, req_adr);
}

static uint32 mmu_block_read_half_word(void *data, uint32 adr,
				       struct profile *profile,
				       uint32 req_adr) {
  struct mmu_block *b = data;
  if ((b->cache->settings & 6) && b->cache_enabled) {
    uint8 *p = cache_read(b, adr, profile, req_adr);
    return get_half_word(p);
  } else
    return memory_read_half_word(b->memory,
				 b->phys_adr + (adr & ((1 << 21) - 1)),
				 profile, req_adr);
}

static uint32 mmu_block_read_word(void *data, uint32 adr,
				  struct profile *profile,
				  uint32 req_adr) {
  struct mmu_block *b = data;
  if ((b->cache->settings & 6) && b->cache_enabled) {
    uint8 *p = cache_read(b, adr, profile, req_adr);
    return get_word(p);
  } else
    return memory_read_word(b->memory, b->phys_adr + (adr & ((1 << 21) - 1)),
			    profile, req_adr);
}

static void mmu_block_write(void *data, uint32 adr, uint8 val,
			    struct profile *profile, uint32 req_adr) {
  struct mmu_block *b = data;
  uint8 *p = cache_write(b, adr, profile, req_adr);
  if (p)
    *p = val;
  else
    memory_write(b->memory, b->phys_adr + (adr & ((1 << 21) - 1)), val,
		 profile, req_adr);
}

static void mmu_block_write_half_word(void *data, uint32 adr, uint32 val,
			    struct profile *profile, uint32 req_adr) {
  struct mmu_block *b = data;
  uint8 *p = cache_write(b, adr, profile, req_adr);
  if (p)
    put_half_word(p, val);
  else
    memory_write_half_word(b->memory,
			   b->phys_adr + (adr & ((1 << 21) - 1)), val);
}

static void mmu_block_write_word(void *data, uint32 adr, uint32 val,
			    struct profile *profile, uint32 req_adr) {
  struct mmu_block *b = data;
  uint8 *p = cache_write(b, adr, profile, req_adr);
  if (p)
    put_word(p, val);
  else
    memory_write_word(b->memory, b->phys_adr + (adr & ((1 << 21) - 1)), val,
		      profile, req_adr);
}

static void mmu_set_mapping(struct hardware *hardware,
			    uint32 dest, uint32 src) {
  hardware->mmu[dest >> 21].phys_block = src >> 21;
  hardware->mmu[dest >> 21].phys_adr = src;
}

static uint32 mmu_read(struct hardware *hardware, uint32 adr) {
  if (adr >= 0x18 && adr < 0x58)
    return hardware->mmu[(adr - 0x18) / 4].phys_block;
  else if (adr == 4)
    return hardware->cache.settings;
  else {
    logmsg(LOG_MEMORY, 5, "mmu: read from %x\n", adr);
    return 0;
  }
}

static void mmu_write(struct hardware *hardware, uint32 adr, uint32 val) {
  if (adr >= 0x18 && adr < 0x58)
    mmu_set_mapping(hardware, (adr - 0x18) << 19, (val & 0x7ff) << 21);
  else if (adr == 4)
    hardware->cache.settings = val;
  else if (adr == 8) {
    int i;
    for (i = 0; i < 16; i++)
      hardware->mmu[i].cache_enabled = (val >> i) & 1;
  } else
    logmsg(LOG_MEMORY, 5, "mmu: write %x to register %x\n", val, adr);
}

void initialize_mmu(struct hardware *hardware, struct memory *phys_memory,
		    struct memory *memory) {
  int i;
  hardware->memory = memory;
  hardware->phys_memory = phys_memory;
  memory_copy_configuration(memory, 0, 0xffffffff, phys_memory, 0);
  for (i = 0; i < 128; i++) {
    hardware->cache.line[i].lru = 0;
    hardware->cache.line[i].addr[0] = 0xff000000;
    hardware->cache.line[i].addr[1] = 0xff000000;
  }
  hardware->cache.settings = 0;
  hardware->cache.block = hardware->mmu;
  for (i = 0; i < 16; i++) {
    hardware->mmu[i].phys_block = i;
    hardware->mmu[i].phys_adr = i * 0x200000;
    hardware->mmu[i].memory = phys_memory;
    hardware->mmu[i].cache = &hardware->cache;
    hardware->mmu[i].cache_enabled = 0;
    set_area(memory, 0x200000 * i, 0x200000 * i + 0x1fffff,
	     mmu_block_read, mmu_block_write,
	     mmu_block_read_half_word, mmu_block_write_half_word,
	     mmu_block_read_word, mmu_block_write_word,
	     &(hardware->mmu[i]));
  }
  mmu_set_mapping(hardware, 0, 0x10400000);
}

static void dma_add_event(struct dma_channel *ch) {
  ch->event.tick = tick + 16 * 500;
  add_event(&(ch->event));
}

static void dma_event(struct event *e) {
  int size, i;
  struct dma_channel *ch = e->data;

  for (i = 0; i < 16 && ch->enable; i++) {
    int size = 2 - bits(ch->config, 10, 2);
    uint32 s1 = (bits(ch->config, 5, 5) == 0) ?
      ch->src + (ch->count << size) : ch->src;
    uint32 d1 = (bits(ch->config, 0, 5) == 0) ?
      ch->dest + (ch->count << size) : ch->dest;

    switch (size) {
    case 0:
      memory_write(ch->dma->memory, d1,
		   memory_read(ch->dma->memory, s1, NULL, 0),
		   NULL, 0);
      break;
    case 1:
      memory_write_half_word(ch->dma->memory, d1,
			     memory_read_half_word(ch->dma->memory, s1,
						   NULL, 0),
			     NULL, 0);
      break;
    case 2:
      logmsg(LOG_MEMORY, 5, "transferring %x -> %x\n", s1, d1);
      memory_write_word(ch->dma->memory, d1,
			memory_read_word(ch->dma->memory, s1, NULL, 0),
			NULL, 0);
      break;
    }

    if (ch->count == ch->len / 2 &&
	!(ch->dma->inten & (1 << (2 * ch->n + 1)))) {
      ch->dma->intstat |= 1 << (2 * ch->n + 1);
      set_irq_request(ch->dma->hardware, 0x1c);
    }
  
    if (ch->count == ch->len) {
      if (!(ch->dma->inten & (1 << (2 * ch->n)))) {
	ch->dma->intstat |= 1 << (2 * ch->n);
	set_irq_request(ch->dma->hardware, 0x1c);
      }
      if (bits(ch->config, 18, 1))
	ch->count = 0;
      else
	ch->enable = 0;
      if (bits(ch->config, 17, 1)) {
	struct dma_channel *pair =
	  &(ch->dma->channel[bits(ch->config, 13, 3)]);
	/* probably it's ok to put the same channel as paired, but
	   it's not very useful */
	if (!pair->enable && pair != ch) {
	  pair->enable = 1;
	  dma_add_event(pair);
	}
      }
    } else
      ch->count++;
  }

  if (ch->enable)
    dma_add_event(ch);
}

static uint32 dma_read(struct dma_ctrl *dma, uint32 adr) {
  switch (adr) {
  case 0x400:
    {
      uint32 r = 0;
      int i;
      logmsg(LOG_MEMORY, 5, "dma: read from enable\n");
      for (i = 0; i < 8; i++)
	if (dma->channel[i].enable) r |= 1 << i;
      return r;
    }
  case 0x404:
    logmsg(LOG_MEMORY, 5, "dma: read from intstat\n");
    return dma->intstat;
  case 0x408:
    logmsg(LOG_MEMORY, 5, "dma: read from inten\n");
    return dma->inten;
  default:
    if (adr >= 0 && adr <= 0x100) {
      int ch = adr >> 5;
      logmsg(LOG_MEMORY, 5, "dma(%d): read from %x\n", ch, adr);
      switch (adr & 0x1f) {
      case 0:
	return dma->channel[ch].src;
      case 4:
	return dma->channel[ch].dest;
      case 8:
	return dma->channel[ch].len;
      case 0xc:
	return dma->channel[ch].config;
      case 0x10:
	return dma->channel[ch].enable;
      case 0x1c:
	return dma->channel[ch].count;
      default:
	return 0;
      }
    } else {
      logmsg(LOG_MEMORY, 5, "dma: read from %x\n", adr);
      return 0;
    }
  }
}

static void dma_write_enable(struct dma_channel *ch, int val) {
  if (!ch->enable && val) {
    ch->enable = 1;
    add_event(&(ch->event));
  } else if (ch->enable && !val) {
    remove_event(&(ch->event));
    ch->enable = 0;
  }
}

static void dma_write(struct dma_ctrl *dma, uint32 adr, uint32 val) {
  int i;
  switch (adr) {
  case 0x400:
    logmsg(LOG_MEMORY, 5, "dma: write %x to enable\n", val);
    for (i = 0; i < 8; i++)
      dma_write_enable(&(dma->channel[i]), (val >> i) & 1);
    break;
  case 0x404:
    logmsg(LOG_MEMORY, 5, "dma: write %x to intstat\n", val);
    dma->intstat &= ~val;
    break;
  case 0x408:
    logmsg(LOG_MEMORY, 5, "dma: write %x to inten\n", val);
    dma->inten = val;
    break;
  case 0x410:
    logmsg(LOG_MEMORY, 5, "dma: write to softint\n", val);
    if (!(dma->inten & (1 << 30))) {
      dma->intstat |= (1 << 30);
      set_irq_request(dma->hardware, 0x1c);
    }
    break;
  default:
    if (adr >= 0x200 && adr < 0x280)
      adr = ((adr & 0x70) << 1) | (adr & 0xf);
    if (adr >= 0 && adr < 0x100) {
      int chn = adr >> 5;
      struct dma_channel *ch = &(dma->channel[chn]);
      logmsg(LOG_MEMORY, 5, "dma(%d): write %x to %x\n", chn, val, adr);
      switch (adr & 0x1f) {
      case 0:
	ch->src = val;
	break;
      case 4:
	ch->dest = val;
	break;
      case 8:
	ch->len = val & 0xfff;
	break;
      case 0xc:
	ch->config = val;
	break;
      case 0x10:
	dma_write_enable(ch, val & 1);
	break;
      case 0x1c:
	ch->count = 0;
	break;
      }
    } else
      logmsg(LOG_MEMORY, 5, "dma: write %x to %x\n", val, adr);
  }
}

static void emu_write(struct hardware *hardware, uint32 adr, uint32 val) {
  switch (adr & 0xff) {
  case 0:
    hardware->log_buf[hardware->log_ptr++] = val;
    if (val == 0 || hardware->log_ptr >= sizeof(hardware->log_buf) - 1) {
      hardware->log_buf[sizeof(hardware->log_buf) - 1] = 0;
      logmsg(LOG_GENERAL, 5, "log: %s\n", hardware->log_buf);
      fprintf(stderr, "%s\n", hardware->log_buf);
      hardware->log_ptr = 0;
    }
    break;
  }
}

static uint8 hardware_flash_read(void *data, uint32 adr,
				 struct profile *profile, uint32 req_adr) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1) {
    logmsg(LOG_FLASH, 5, "flash: no chip selected during read from %x\n", adr);
    return ~0;
  }

  flash = hardware->flash[chip];
  if (!flash)
    return ~0;

  return flash_read(flash, adr & 0xf);
}

static void hardware_flash_write(void *data, uint32 adr, uint8 val,
				 struct profile *profile, uint32 req_adr) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1) {
    logmsg(LOG_FLASH, 5, "flash: no chip selected during write %x to %x\n", val, adr);
    return;
  }

  flash = hardware->flash[chip];
  if (!flash)
    return;

  flash_write(flash, adr & 0xf, val);
}

static int cgu_log_counter = 0;

static uint32 hardware_read_word(void *data, uint32 adr,
			    struct profile *profile, uint32 req_adr) {
  struct hardware *hardware = data;

  {
    if (cgu_log_counter >= 10000) {
      cgu_log_counter = 0;
      clocks_log_state(&hardware->clocks);
    } else
      cgu_log_counter++;
  }

  switch ((adr >> 8) & 0xfffc) {
  case 0x30:
    return gpio_read(data, adr);
  case 0x40:
  case 0x44:
  case 0x48:
  case 0x4c:
    return clocks_read(&(hardware->clocks), adr & 0xfff);
  case 0x80:
    /* emc_read */
    return 0;
  case 0x1040:
    return lcd_read(hardware->lcd, adr);
  case 0x1048:
  case 0x104c:
    return dma_read(&(hardware->dma_ctrl), adr & 0x7ff);
  case 0x1050:
    return mmu_read(hardware, adr & 0xff);
  case 0x3000:
  case 0x3004:
    return read_803(data, adr);
  case 0x24:
    return adc_read(hardware, adr & 0xff);
  case 0x200:
    timer_read(hardware, adr & 0x7ff);
    break;
  case 0x204:
    timer_read(hardware, adr & 0x7ff);
    break;
  case 0x208:
    return i2c_read(hardware, adr & 0xff);
  case 0x2000:
    return audio_read(&(hardware->audio), adr & 0xfff);
  default:
    logmsg(LOG_MEMORY, 10, "read from %8x\n", adr);
    return 0xffffffff;
  }
}

static uint32 hardware_read_half_word(void *data, uint32 adr,
				      struct profile *profile,
				      uint32 req_adr) {
  struct hardware *hardware = data;

  switch ((adr >> 8) & 0xfffc) {
  case 0x2000:
    return audio_read(&(hardware->audio), adr & 0xfff);
  default:
    logmsg(LOG_MEMORY, 10, "read half word from %8x\n", adr);
    return 0xffff;
  }
}

static void hardware_write_word(void *data, uint32 adr, uint32 val,
			    struct profile *profile, uint32 req_adr) {
  struct hardware *hardware = data;

  switch ((adr >> 8) & 0xfffc) {
  case 0x30:
    gpio_write(data, adr, val);
    break;
  case 0x40:
  case 0x44:
  case 0x48:
  case 0x4c:
    clocks_write(&(hardware->clocks), adr & 0xfff, val);
    break;
  case 0x1040:
    lcd_write(hardware->lcd, adr, val);
    break;
  case 0x1048:
  case 0x104c:
    dma_write(&(hardware->dma_ctrl), adr & 0x7ff, val);
    break;
  case 0x1050:
    mmu_write(hardware, adr & 0xff, val);
    break;
  case 0x3000:
  case 0x3004:
    write_803(data, adr, val);
    break;
  case 0x24:
    adc_write(hardware, adr & 0xff, val);
    break;
  case 0x208:
    i2c_write(hardware, adr & 0xff, val);
    break;
  case 0x200:
    timer_write(hardware, adr & 0x7ff, val);
    break;
  case 0x204:
    timer_write(hardware, adr & 0x7ff, val);
    break;
  case 0x2000:
    audio_write(&(hardware->audio), adr & 0xfff, val);
    break;
  case 0x3f00:
    emu_write(hardware, adr & 0xff, val);
    break;
  default:
    logmsg(LOG_MEMORY, 10, "write %8x to %8x (pc=%8x)\n", val, adr, cpu->reg[15] - 4);
  }
}

static void hardware_write_half_word(void *data, uint32 adr, uint32 val,
				     struct profile *profile, uint32 req_adr) {
  struct hardware *hardware = data;

  switch ((adr >> 8) & 0xfffc) {
  case 0x2000:
    audio_write(&(hardware->audio), adr & 0xfff, val);
    break;
  default:
    logmsg(LOG_MEMORY, 10, "write half word %8x to %8x (pc=%8x)\n",
	   val, adr, cpu->reg[15] - 4);
  }
}

void install_hardware(struct hardware *hardware,
		      struct memory *memory) {
  set_area(memory, 0x80000000, 0x803fffff,
	   read_void, write_void,
	   hardware_read_half_word, hardware_write_half_word,
	   hardware_read_word, hardware_write_word, hardware);
  set_area(memory, 0x20000000, 0x200fffff,
	   prog_flash_read_byte, write_void,
	   prog_flash_read_half_word, prog_flash_write_half_word,
	   prog_flash_read_word, write_word_void,
	   hardware->prog_flash);
  set_area(memory, 0x28000000, 0x2800ffff,
	   hardware_flash_read, hardware_flash_write,
	   read_half_word_void, write_half_word_void,
	   read_word_void, write_word_void, hardware);
  /*
  set_area(memory, 0x24100000, 0x2410ffff,
	   usb_read, usb_write,
	   usb_read_half_word, usb_write_half_word,
	   usb_read_word, usb_write_word, hardware->usb);
  */
}

struct prog_flash *create_prog_flash(char *fname) {
  struct prog_flash *flash;

  flash = xmalloc(sizeof(struct prog_flash));
  if (!flash)
    return NULL;

  flash->state = 0;

  flash->content = xmalloc(1024 * 1024);
  if (!flash->content) {
    free(flash);
    return NULL;
  }
  memset(flash->content, 0xff, 1024 * 1024);

  flash->fp = fopen(fname, "r+");
  if (!flash->fp) {
    error_sys("%s: ", fname);
    free(flash->content);
    free(flash);
    return NULL;
  }

  fread(flash->content, 1, 1024 * 1024, flash->fp);
  return flash;
}

void free_prog_flash(struct prog_flash *flash) {
  if (flash) {
    if (flash->fp)
      fclose(flash->fp);
    if (flash->content)
      free(flash->content);
    free(flash);
  }
}

void audio_reset(struct audio *audio) {
  audio->dacctrl = audio->dacset = 0;
  audio->siocr = 0x80;

  audio->out_file[0] = fopen("left.out", "wb");
  audio->out_file[1] = fopen("right.out", "wb");
}

void clocks_reset(struct clocks *clocks) {
  int i;
  for (i = 0; i < 12; i++) {
    clocks->base_scr[i] = 0;
    clocks->base_fs1[i] = 0;
    clocks->base_fs2[i] = 0;
    clocks->base_ssr[i] = 0;
  }
  for (i = 0; i < 73; i++)
    clocks->clk_pcr[i] = 0;
  for (i = 0; i < 67; i++)
    clocks->clk_esr[i] = 0;
  for (i = 0; i < 3; i++)
    clocks->base_bcr[i] = 0;
  for (i = 0; i < 18; i++)
    clocks->base_fdc[i] = 0;
}

void hardware_reset(struct hardware *hardware, struct memory *phys_memory) {
  int i;
  tick = 0;
  for (i = 0; i < 8; i++) {
    hardware->gpio_out[i] = ~0;
    hardware->gpio_dir[i] = 0;
  }
  hardware->gpio_out[2] &= ~1; /* power off bit */
  
  for (i = 0; i < 2; i++) {
    hardware->timer[i].ctrl = 0;
    hardware->timer[i].index = i;
    hardware->timer[i].hardware = hardware;
    hardware->timer[i].active = 0;
    hardware->timer[i].event.f = timer_event;
    hardware->timer[i].event.data = &(hardware->timer[i]);
  }
  
  hardware->dma_ctrl.hardware = hardware;
  hardware->dma_ctrl.memory = phys_memory;
  hardware->dma_ctrl.inten = 0xffff;
  hardware->dma_ctrl.intstat = 0;
  for (i = 0; i < 8; i++) {
    hardware->dma_ctrl.channel[i].event.f = dma_event;
    hardware->dma_ctrl.channel[i].event.data =
      &(hardware->dma_ctrl.channel[i]);
    hardware->dma_ctrl.channel[i].n = i;
    hardware->dma_ctrl.channel[i].dma = &(hardware->dma_ctrl);
  }

  hardware->irq_request = 0;

  audio_reset(&(hardware->audio));
  clocks_reset(&(hardware->clocks));

  hardware->log_ptr = 0;
}

static int usb_get_connected(void *data, int port, int bit) {
  return ((struct hardware *)data)->usb_connected;
}

static int get_hold(void *data, int port, int bit) {
  return !ui_get_key(9);
}

static int get_port6b0(void *data, int port, int bit) {
  return 0;
}

static void set_backlight(void *data, int port, int bit, int v) {
  logmsg(LOG_GPIO, 5, "backlight: set to %d\n", v);
}

static void set_rtc_i2c(void *data, int port, int bit, int v) {
  logmsg(LOG_GPIO, 5, "rtc: i2c %s set to %d\n", bit ? "scl" : "sda", v);
}

static void set_power(void *data, int port, int bit, int v) {
  if (v) {
    logmsg(LOG_GPIO, 5, "gpio: power off");
    cpu->stop = CPU_STOP_QUIT;
  }
}

static void set_unknown_port(void *data, int port, int bit, int v) {
  logmsg(LOG_GPIO, 5, "gpio: bit %d/%d set to %d\n", port, bit, v);
}

void hardware_connect_flash(struct hardware *hardware,
			    int n, struct flash *flash) {
  hardware->flash[n] = flash;
}


struct hardware *create_hardware(void) {
  int i, j;
  struct hardware *hardware = xmalloc(sizeof(struct hardware));
  if (!hardware) return NULL;

  hardware->prog_flash = create_prog_flash(prog_flash_file);
  if (!hardware->prog_flash)
    goto no_prog_flash;

  for (i = 0; i < 4; i++)
    hardware->flash[i] = NULL;

  for (i = 0; i < 8; i++)
    for (j = 0; j < 32; j++) {
      hardware->gpio_handler[i][j].get = NULL;
      hardware->gpio_handler[i][j].set = NULL;
    }

  gpio_install_handler(hardware, 6, 3,
		       hardware, flash_get_busy, NULL);

  gpio_install_handler(hardware, 7, 0,
		       hardware, usb_get_connected, NULL);

  gpio_install_handler(hardware, 5, 2,
		       hardware, get_hold, NULL);

  gpio_install_handler(hardware, 6, 0,
		       hardware, get_port6b0, set_rtc_i2c);
  gpio_install_handler(hardware, 6, 1,
		       hardware, NULL, set_rtc_i2c);

  gpio_install_handler(hardware, 3, 0,
		       hardware, NULL, set_backlight);

  gpio_install_handler(hardware, 2, 0,
		       hardware, NULL, set_power);
  gpio_install_handler(hardware, 1, 9,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 19,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 11,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 3, 2,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 8,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 3, 6,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 16,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 5, 0,
		       hardware, NULL, set_unknown_port);

  hardware->lcd = create_lcd();
  if (!hardware->lcd)
    goto no_lcd;


  hardware->usb = create_usb();
  if (!hardware->usb)
    goto no_usb;
  hardware->usb_connected = 0;

  return hardware;

 no_usb:
  free_lcd(hardware->lcd);
 no_lcd:
  fclose(hardware->flash[0]->fp);
 no_fp:
  free(hardware->flash[0]);
 no_flash:
  free_prog_flash(hardware->prog_flash);
 no_prog_flash:
  free(hardware);
 err:
  return NULL;

}
