/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <errno.h>

#include "emuriver.h"
#include "memory.h"
#include "cpu.h"
#include "hardware.h"
#include "trace.h"
#include "trainer.h"
#include "gdb.h"
#include "ui.h"
#include "flash.h"
#include "profile.h"

struct memory *memory, *phys_memory;
struct cpu *cpu;
struct hardware *hardware;

struct mem_block *iram, *dram, *xram;

char *program_name;

#define MODEL_790 0
#define MODEL_795 1
#define MODEL_799 2

int option_gdb = 0;
int option_disable_ui = 0;
int option_profile = 0;
int option_model = MODEL_790;
int option_extra_memory = 0;
int option_startup_hold_mode = 0;
int option_startup_hold_on = 0;

char *log_file = NULL;
char *flash_file = "flash.img";
char *prog_flash_file = "prog_flash.bin";
char *int_flash_file = "ifp1.bin";
char *rockbox_file = NULL;

struct option options[] = {
  {"log", required_argument, NULL, 'l'},
  {"gdb", no_argument, NULL, 'g'},
  {"disable-ui", no_argument, NULL, 'n'},
  {"profile", no_argument, NULL, 'P'},
  {"flash-image", required_argument, NULL, 'f'},
  {"prog-flash-image", required_argument, NULL, 'p'},
  {"int-flash-image", required_argument, NULL, 'i'},
  {"rockbox", required_argument, NULL, 'r'},
  {"model", required_argument, NULL, 'm'},
  {"extra-memory", no_argument, NULL, 'x'},
  {0, 0, 0, 0}
};

void error(const char *format, ...) {
  va_list ap;
  fprintf(stderr, "%s: ", program_name);
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fputc('\n', stderr);
}

void error_sys(const char *format, ...) {
  va_list ap;
  fprintf(stderr, "%s: ", program_name);
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  fprintf(stderr, "%s\n", strerror(errno));
}

void *xmalloc(size_t size) {
  void *p = malloc(size);
  if (!p)
    error("out of memory");
  return p;
}

int set_iriver_memory(struct memory *memory) {
  FILE *fp;
  iram = create_memory_area_ram(memory, 0x400000, 0x10000,
				PROFILE_CTR_IRAM_ACCESSES);
  dram = create_memory_area_ram(memory, 0x24000000, 0x100000,
				PROFILE_CTR_SRAM_ACCESSES);
  if (option_extra_memory)
    xram = create_memory_area_ram(memory, 0x30000000, 0x800000,
				  PROFILE_CTR_SRAM_ACCESSES);
  else
    xram = NULL;

  if (!iram || !dram || (option_extra_memory && !xram))
    return -1;
  if (!create_memory_area_rom(memory, 0x10400000, 0x80000,
			      int_flash_file))
    return -1;

  if (rockbox_file) {
    fp = fopen(rockbox_file, "rb");
    if (!fp) {
      error_sys("%s: ", rockbox_file);
      return -1;
    }
    fread(option_extra_memory ? xram->mem : dram->mem, 1, 0x80000, fp);
    fclose(fp);
  }
  return 0;
}

int set_mpio_memory(struct memory *memory) {
  iram = create_memory_area_ram(memory, 0x20010000, 0x20000, -1);
  if (!iram)
    return -1;
  return 0;
}

int save_ram(struct mem_block *mem, char *file) {
  FILE *fp;
  fp = fopen(file, "wb");
  if (!fp) return -1;
  fwrite(mem->mem, 1, 0x10000, fp);
  fclose(fp);

  return 0;
}

int load_binary(char *file, void *ptr, size_t len) {
  FILE *fp;
  fp = fopen(file, "rb");
  if (!fp) return -1;
  fread(ptr, 1, len, fp);
  fclose(fp);
}

int init_iriver(struct cpu *cpu, struct memory *memory, int upgrade) {
  if (set_iriver_memory(memory) < 0)
    return -1;

  if (upgrade) {
    load_binary("fw_upgrade.bin", iram->mem + 0x80, 0x10000);
    cpu->reg[15] = 0x000080;
  } else if (rockbox_file)
    cpu->reg[15] = option_extra_memory ? 0x30000000 : 0x24000000;
  else
    cpu->reg[15] = 0x10400000;
  return 0;
}

void init_mpio(struct cpu *cpu, struct memory *memory, int upgrade) {
  set_mpio_memory(memory);
  if (upgrade) {
    load_binary("mpio_upgrade.bin", iram->mem + 0x80, 0x10000);
    cpu->reg[15] = 0x000080;
  } else
    cpu->reg[15] = 0x800000;
}

int go(int n_cycles, int (*f)(void), int freq) {
  int i, stop, r;

  i = 0;
  while (n_cycles == 0 || i < n_cycles) {
    stop = cpu_go(cpu, 1);
    tick++;

    handle_events(tick);

    /*
    if ((tick % 1000000) == 5000) {
      cpu_irq(cpu);
      set_irq_no(0x1c);
    }
    */

    if (stop != CPU_STOP_NORMAL)
      return stop;

    if (f && (tick % freq) == 0) {
      r = (*f)();
      if (r)
	return r;
    }

    if (!option_disable_ui && (tick % 100000) == 0) {
      ui_refresh(hardware->lcd);
      if (ui_get_key(0)) return CPU_STOP_QUIT;
    }

    i++;
  }

  return CPU_STOP_NORMAL;
}

int parse_options(int argc, char *argv[]) {
  int c, index;
  FILE *log_fp;

  while (1) {
    c = getopt_long(argc, argv, "l:gnPf:p:i:m:x", options, &index);
    if (c == -1)
      break;

    switch (c) {
    case '?':
      return -1;
    case 'l':
      log_file = optarg;
      break;
    case 'g':
      option_gdb = 1;
      break;
    case 'n':
      option_disable_ui = 1;
      break;
    case 'P':
      option_profile = 1;
      break;
    case 'f':
      flash_file = optarg;
      break;
    case 'p':
      prog_flash_file = optarg;
      break;
    case 'i':
      int_flash_file = optarg;
      break;
    case 'r':
      rockbox_file = optarg;
      break;
    case 'm':
      if (!strcmp(optarg, "790"))
	option_model = MODEL_790;
      else if (!strcmp(optarg, "795"))
	option_model = MODEL_795;
      else if (!strcmp(optarg, "799"))
	option_model = MODEL_799;
      else {
	fprintf(stderr, "%s: unknown model %s\n", argv[0], optarg);
	return -1;
      }
      break;
    case 'x':
      option_extra_memory = 1;
      break;
    default:
      abort();
    }
  }

  if (log_file) {
    log_fp = fopen(log_file, "w");
    if (log_fp)
      log_set_fp(log_fp);
  }
}

static int profile_counters[] = {
  PROFILE_CTR_CPU_CYCLES,
  PROFILE_CTR_IRAM_ACCESSES,
  PROFILE_CTR_SRAM_ACCESSES,
  PROFILE_CTR_CACHE_HITS,
  PROFILE_CTR_CACHE_MISSES,
  PROFILE_CTR_CACHE_FLUSHES,
  -1
};

int main(int argc, char *argv[]) {
  int upgrade = 1;
  struct flash *flash;
  struct profile *profile;
  int i;
  
  program_name = argv[0];
  if (parse_options(argc, argv) == -1)
    return 1;

  init_events();

  if (option_profile) {
    profile = create_profile();
    if (!profile)
      return 1;
    
    for (i = 0; profile_counters[i] >= 0; i++) {
      if (profile_add_counter(profile, profile_counters[i],
			      0xc00000, 0xcfffff) < 0)
	return 1;
      if (profile_add_counter(profile, profile_counters[i],
			      0x400000, 0x40ffff) < 0)
	return 1;
    }
  } else
    profile = NULL;
  
  memory = create_memory();
  if (!memory)
    return 1;

  phys_memory = create_memory();
  if (!phys_memory)
    return 1;

  cpu = create_cpu(memory, profile);
  if (!cpu)
    return 1;

  hardware = create_hardware();
  if (!hardware)
    return 1;

  switch (option_model) {
  default:
  case MODEL_790:
    flash = create_flash(FLASH_256_8, flash_file, 0);
    if (!flash)
      return 1;
    hardware_connect_flash(hardware, 0, flash);
    break;

  case MODEL_799:
    flash = create_flash(FLASH_512_8, flash_file, 0);
    if (!flash)
      return 1;
    hardware_connect_flash(hardware, 0, flash);
    flash = create_flash(FLASH_512_8, flash_file, 2112 * (1 << 18));
    if (!flash)
      return 1;
    hardware_connect_flash(hardware, 1, flash);
    break;
  }

  cpu_set_irq_p(cpu, check_irq, hardware);

  hardware_reset(hardware, phys_memory);
  install_hardware(hardware, phys_memory);

  if (init_iriver(cpu, phys_memory, 0) < 0)
    return 1;

  initialize_mmu(hardware, phys_memory, memory);

  trace_detect_firmware(memory);
  trace_init();
  if (rockbox_file)
    stack_tracing = 0;

  trainer_init(iram->mem);

  if (option_disable_ui || !ui_init()) {
    if (option_gdb)
      gdb_target();
    else
      go(0, NULL, 0);

    if (!option_disable_ui)
      ui_cleanup();
  }

  if (option_profile) {
    FILE *fp = fopen("profile.out", "wt");
    if (!fp) return 1;
    profile_dump(profile, fp);
    fclose(fp);
  }

  return 0;
}
