/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>

#include "trace.h"
#include "cpu.h"
#include "memory.h"
#include "log.h"

static int link_stack[10000];
static int nesting_level = 0;
static int quiet = 0;
static struct function *functions;
static int n_functions;
static uint32 last_pc, last_func;

int stack_tracing;

static struct function functions_empty[] = {
  {0, NULL, 0, 0}
};

void trace_init(void) {
  int i;
  i = 0;
  while (functions[i].adr != 0) i++;
  n_functions = i;
  last_pc = 0xdeadcafe;
  stack_tracing = 1;
}

static int trace_fw_strcmp(struct memory *mem, uint32 adr, char *s) {
  while (*s) {
    if (memory_read(mem, adr, NULL, 0) != *s) return 0;
    s++;
    adr++;
  }
  return 1;
}

void trace_detect_firmware(struct memory *mem) {
  functions = functions_empty;
}

static int print_backtrace(char *s) {
  int i, p, q;
  sprintf(s, "[");
  p = 1;
  for (i = 0; i < nesting_level; i++) {
    sprintf(s + p, "%x%c%n", link_stack[i],
	    i == nesting_level - 1 ? ']' : ' ', &q);
    p += q;
  }
  return p;
}

void trace_bl(struct cpu *cpu, int backtrace) {
  char tmp[1000]; /* TODO may overflow */
  int i;
  if (quiet)
    return;

  if (stack_tracing) {
    if (nesting_level > 30) return;

    link_stack[nesting_level++] = cpu->reg[14];

    if (backtrace) {
      int p = 0;
      int q;
      sprintf(tmp + p,
	      "bl to %8x %n", cpu->reg[15], &q);
      p += q;
      p += print_backtrace(tmp + p);
      tmp[p] = '\n';
      tmp[p + 1] = 0;
    } else {
      sprintf(tmp + nesting_level,
	      "bl to %8x (from %8x)\n", cpu->reg[15], cpu->reg[14]);
    }
  
    logmsg(LOG_TRACE, 10, tmp);
    log_inc_indent();
  } else
    logmsg(LOG_TRACE, 10, "bl to %8x (from %8x)\n", cpu->reg[15], cpu->reg[14]);
}

static void function_entry(struct cpu *cpu, struct function *f) {
  int i, p, q, logon;
  char tmp[200];

  logon = !quiet && log_check(LOG_TRACE, 5);
  if (logon) {
    sprintf(tmp, "function %s(%n", f->name, &p);
    for (i = 0; i < f->n_args; i++) {
      uint32 val;

      if (i < 4)
	val = cpu->reg[i];
      else
	val = memory_read_word(cpu->mem, cpu->reg[13] + (i - 4) * 4, NULL, 0);

      if (i > 0) {
	strcpy(tmp + p, ", ");
	p += 2;
      }
      sprintf(tmp + p, "0x%x%n", val, &q);
      p += q;
    }
  }

  if (f->action == 1) {
    if (logon)
      sprintf(tmp + p, ") (ignored)\n");
    cpu->reg[15] = cpu->reg[14];
  } else if (f->action == 2) {
    if (logon)
      sprintf(tmp + p, ") (folded)\n");
    quiet = 1;
  } else
    if (logon)
      sprintf(tmp + p, ")\n");
  if (logon)
    logmsg(LOG_TRACE, 5, tmp);
}

void trace(struct cpu *cpu) {
  int i;

  /*
  if (cpu->reg[15] >= 0x83e5e8 && cpu->reg[15] <= 0x83e69c) {
    logmsg(LOG_TRACE, 5, "in div64: cpsr=%8x\n", cpu->cpsr);
    for (i = 0; i < 4; i++)
      logmsg(LOG_TRACE, 5, "r%02d=%8x r%02d=%8x r%02d=%8x r%02d=%8x\n",
	     i * 4, cpu->reg[i * 4], i * 4 + 1, cpu->reg[i * 4 + 1],
	     i * 4 + 2, cpu->reg[i * 4 + 2], i * 4 + 3, cpu->reg[i * 4 + 3]);
  }
  */

  /*
  if (cpu->reg[15] == 0x40da24 || cpu->reg[15] == 0x40e0b8) {
    uint32 start, n;
    static uint32 a, n1;
    if (cpu->reg[15] == 0x40da24) {
      start = a = cpu->reg[1];
      n = n1 = cpu->reg[0];
    }else {
      start = a;
      n = n1;
    }

    logmsg(LOG_TRACE, 5, "mdct %s:\n", cpu->reg[15] == 0x40da24 ? "in" : "out");
    for (i = 0; i < n; i += 8)
      logmsg(LOG_TRACE, 5, "m: %8x %8x %8x %8x %8x %8x %8x %8x\n",
	     memory_read_word(cpu->mem, start + i + 0),
	     memory_read_word(cpu->mem, start + i + 4),
	     memory_read_word(cpu->mem, start + i + 8),
	     memory_read_word(cpu->mem, start + i + 12),
	     memory_read_word(cpu->mem, start + i + 16),
	     memory_read_word(cpu->mem, start + i + 20),
	     memory_read_word(cpu->mem, start + i + 24),
	     memory_read_word(cpu->mem, start + i + 28));
  }
  */
  
  if (cpu->reg[14] == last_pc + 4)
    trace_bl(cpu, 1);
  last_pc = cpu->reg[15];

  int a, b;
  if (last_pc + 4 == cpu->reg[15]) {
    a = last_func;
    b = last_func + 2;
    if (b > n_functions) b = n_functions;
  } else {
    a = 0;
    b = n_functions;
  }
  while (b - a > 1) {
    int m = (a + b) / 2;
    if (functions[m].adr <= cpu->reg[15])
      a = m;
    else
      b = m;
  }
  last_func = a;
  if (functions[a].adr == cpu->reg[15])
    function_entry(cpu, functions + a);

  if (stack_tracing) {
    while (nesting_level > 0
	   && link_stack[nesting_level - 1] == cpu->reg[15]) {
      quiet = 0;
      nesting_level--;
      log_dec_indent();
    }
  }
}
